@@ -16,21 +16,25 @@ in a **Docker** container listening on port 1883 for authenticated connections.
 import json
 import logging
 import os
+import threading
 from atexit import register as on_exit
-from enum import Enum, IntEnum
-from socket import timeout  # : for Python < 3.10 compatibility vs TimeoutError
-from threading import enumerate as enumerate_threads
+from enum import IntEnum
+from socket import gaierror, timeout  # : Python<3.10 vs TimeoutError
 from time import sleep, time
+from typing import Any, Callable
 
 from dotenv import load_dotenv
 from paho.mqtt.client import Client as PahoClient
 from paho.mqtt.client import MQTTMessage as PahoMessage
 
+from fieldedge_utilities.logger import verbose_logging
+from fieldedge_utilities.properties import json_compatible
+
 MQTT_HOST = os.getenv('MQTT_HOST', 'fieldedge-broker')
 MQTT_USER = os.getenv('MQTT_USER')
 MQTT_PASS = os.getenv('MQTT_PASS')
 
-JSON_DUMPS_COMPATIBLE = (dict, list, tuple, str, int, float, Enum, bool)
+__all__ = ['MqttResultCode', 'MqttError', 'MqttClient']
 
 _log = logging.getLogger(__name__)
 
@@ -38,6 +42,7 @@ load_dotenv()
 
 
 class MqttResultCode(IntEnum):
+    """Eclipse Paho MQTT Error Codes."""
     SUCCESS = 0
     ERR_INCORRECT_PROTOCOL = 1
     ERR_INVALID_CLIENT_ID = 2
@@ -61,10 +66,11 @@ class MqttResultCode(IntEnum):
     ERR_UNSUPPORTED_PACKET_TYPE = 20
 
 
-def _get_mqtt_result(rc: int) -> str:
-    if rc in MqttResultCode:
-        return MqttResultCode(rc).name
-    return 'UNKNOWN'
+def _get_mqtt_result(result_code: int) -> str:
+    try:
+        return MqttResultCode(result_code).name
+    except ValueError:
+        return 'UNKNOWN'
 
 
 class MqttError(Exception):
@@ -97,8 +103,7 @@ class MqttClient:
         """Initializes a managed MQTT client.
         
         Args:
-            client_id (str): The unique client ID (default importing module
-                `__name__`)
+            client_id (str): The client ID (default imports module `__name__`)
             on_message (callable): The callback when subscribed messages are
                 received as `topic`(str), `message`(dict|str).
             subscribe_default (str|list[str]): The default subscription(s)
@@ -107,9 +112,12 @@ class MqttClient:
                 attempts if auto_connect is `True`.
             auto_connect (bool): Automatically attempts to connect when created
                 or reconnect after disconnection.
-            **kwargs: include Paho MQTT Client overrides such as:
+            **kwargs: includes advanced feature overrides such as:
             
-            * `on_connect` and/or `on_disconnect`
+            * `client_uid` defaults to True, appends a timestamp to the
+            client_id to avoid being rejected by the host.
+            * `bind_address` to bind to a specific IP
+            * `on_connect`, `on_disconnect`, `on_log` callbacks
             * `host`, `port` (default 1883) and `keepalive` (default 60)
             * `username` and `password`
             * `ca_certs`, `certfile`, `keyfile`
@@ -135,23 +143,22 @@ class MqttClient:
             _log.warning('No on_message specified')
         on_exit(self._cleanup)
         self.on_message = on_message
-        self.on_connect = kwargs.get('on_connect', None)
-        self.on_disconnect = kwargs.get('on_disconnect', None)
+        self.on_connect: 'Callable|None' = kwargs.get('on_connect', None)
+        self.on_disconnect: 'Callable|None' = kwargs.get('on_disconnect', None)
         self._qos = kwargs.get('qos', 0)
-        self._client_id = client_id
+        self._thread_name: str = kwargs.get('thread_name', None)
+        self._client_base_id = client_id
+        self._client_id = None
         self._client_uid = kwargs.get('client_uid', True)
         if self._host.endswith('azure-devices.net'):
-            _log.info('Configuring Azure IoT Hub subscriptions')
+            _log.debug(f'Assuming {client_id} is the Azure IoT Device ID')
             self._client_uid = False
-            subscribe_default = [
-                f'devices/{self.client_id}/messages/devicebound/#',
-                '$iothub/twin/res/#',
-                '$iothub/methods/POST/#',
-            ]
-        if self._client_uid:
-            self.client_id = client_id
-        self._mqtt = PahoClient()
-        self.is_connected = False
+        self.client_id = client_id
+        self._clean_session = kwargs.get('clean_session', True)
+        self._mqtt = PahoClient(clean_session=self._clean_session,
+                                reconnect_on_failure=False)
+        self._connect_timeout = 5
+        self.connect_timeout = int(kwargs.get('connect_timeout', 5))
         self._subscriptions = {}
         self.connect_retry_interval = connect_retry_interval
         self.auto_connect = auto_connect
@@ -163,23 +170,25 @@ class MqttClient:
                 self.subscribe(sub, self._qos)
         if self.auto_connect:
             self.connect()
-    
+
     @property
     def client_id(self):
         return self._client_id
-    
+
     @client_id.setter
-    def client_id(self, id: str):
+    def client_id(self, uid: str):
         if not self._client_uid:
-            self._client_id = id
+            self._client_id = uid
         else:
-            try:
-                if isinstance(int(id.split('_')[1]), int):
-                    # previously made unique, could be a bouncing connection
-                    id = id.split('_')[0]
-            except (ValueError, IndexError):
-                pass   #: new id will be made unique
-            self._client_id = f'{id}_{int(time())}'
+            if uid != self._client_base_id:
+                if verbose_logging('mqtt'):
+                    _log.debug(f'Updating client_id {uid} with new timestamp')
+                uid = self._client_base_id
+            self._client_id = f'{uid}_{int(time())}'
+
+    @property
+    def is_connected(self) -> bool:
+        return self._mqtt.is_connected()
 
     @property
     def subscriptions(self) -> dict:
@@ -196,21 +205,62 @@ class MqttClient:
     def failed_connection_attempts(self) -> int:
         return self._failed_connect_attempts
 
+    @property
+    def on_log(self) -> Callable:
+        return self._mqtt.on_log
+
+    @on_log.setter
+    def on_log(self, callback: Callable):
+        if not isinstance(callback, Callable):
+            raise ValueError('Callback must be a function')
+        self._mqtt.on_log = callback
+
+    @property
+    def connect_timeout(self) -> int:
+        return int(self._mqtt._connect_timeout)
+
+    @connect_timeout.setter
+    def connect_timeout(self, value: 'int|float'):
+        if (not isinstance(value, (int, float)) or
+            not 0 < value <= 120):
+            # invalid value
+            raise ValueError('Connect timeout must be 1..120 seconds')
+        self._connect_timeout = value
+        self._mqtt._connect_timeout = float(value)
+
     def _cleanup(self, *args):
         # TODO: logging raises an error since the log file was closed
-        # for arg in args:
-        #     _log.debug(f'mqtt cleanup called with arg = {arg}')
-        # _log.debug('Terminating MQTT connection')
+        if _vlog():
+            try:
+                for arg in args:
+                    _log.debug(f'mqtt cleanup called with arg = {arg}')
+                _log.debug('Terminating MQTT connection')
+            except:
+                pass
         self._mqtt.user_data_set('terminate')
         self._mqtt.loop_stop()
         self._mqtt.disconnect()
-    
+
+    def _unique_thread_name(self, before_names: 'list[str]') -> str:
+        basename = 'MqttThread'
+        if self._thread_name:
+            basename += f'-{self._thread_name}'
+        name = basename
+        number = 1
+        for name in before_names:
+            if name.startswith(basename):
+                number += 1
+                name = f'{basename}-{number}'
+        return name
+
     def connect(self):
         """Attempts to establish a connection to the broker and re-subscribe."""
         try:
-            _log.debug(f'Attempting MQTT broker connection to {self._host}'
-                       f' as {self._client_id}')
+            if _vlog():
+                _log.debug(f'Attempting MQTT broker connection to {self._host}'
+                           f' as {self.client_id}')
             self._mqtt.reinitialise(client_id=self.client_id)
+            self.connect_timeout = self._connect_timeout
             self._mqtt.user_data_set(None)
             self._mqtt.on_connect = self._mqtt_on_connect
             self._mqtt.on_disconnect = self._mqtt_on_disconnect
@@ -228,24 +278,24 @@ class MqttClient:
                                port=self._port,
                                keepalive=self._keepalive,
                                bind_address=self._bind_address)
-            threads_before = enumerate_threads()
+            threads_before = threading.enumerate()
             self._mqtt.loop_start()
-            threads_after = enumerate_threads()
-            for thread in threads_after:
-                if thread in threads_before:
-                    continue
-                thread.name = 'MqttThread'
-                break
-        except (ConnectionError, timeout, TimeoutError) as err:
+            threads_after = threading.enumerate()
+            new_thread = list(set(threads_after) - set(threads_before))[0]
+            before_names = [t.name for t in threads_before]
+            new_thread.name = self._unique_thread_name(before_names)
+            _log.debug(f'New MQTT client thread: {new_thread.name}')
+            return
+        except (ConnectionError, TimeoutError, gaierror, timeout) as exc:
+            self._mqtt.loop_stop()
             self._failed_connect_attempts += 1
-            if self.connect_retry_interval > 0:
-                _log.warning(f'Unable to connect to {self._host} ({err})'
-                             f' - retrying in {self.connect_retry_interval} s')
-                sleep(self.connect_retry_interval)
-                self.connect()
-            else:
-                _log.warning(f'Failed to connect to {self._host}'
-                             ' but retry disabled - call connect() to retry')
+            _log.error(f'Failed attempt {self._failed_connect_attempts}'
+                       f' to connect to {self._host} ({exc})')
+        # avoid recursing the exception in the stack
+        if self.auto_connect and self.connect_retry_interval > 0:
+            _log.debug(f'Retrying in {self.connect_retry_interval} s')
+            sleep(self.connect_retry_interval)
+            self.connect()
 
     def disconnect(self):
         """Attempts to disconnect from the broker."""
@@ -255,28 +305,28 @@ class MqttClient:
 
     def _mqtt_on_connect(self,
                          client: PahoClient,
-                         userdata: any,
+                         userdata: Any,
                          flags: dict,
-                         rc: int):
+                         reason_code: int):
         """Internal callback re-subscribes on (re)connection."""
         self._failed_connect_attempts = 0
-        if rc == MqttResultCode.SUCCESS:
-            _log.debug(f'Established MQTT connection to {self._host}')
-            if not self.is_connected:
-                for sub in self.subscriptions:
-                    self._mqtt_subscribe(sub, self.subscriptions[sub]['qos'])
-                self.is_connected = True
-            if self.on_connect:
-                self.on_connect(client, userdata, flags, rc)
+        if reason_code == MqttResultCode.SUCCESS:
+            if _vlog():
+                _log.debug(f'Established MQTT connection to {self._host}')
+            for sub, meta in self.subscriptions.items():
+                self._mqtt_subscribe(sub, qos=meta.get('qos', None))
+            if callable(self.on_connect):
+                self.on_connect(client, userdata, flags, reason_code)
         else:
-            _log.error(f'MQTT broker connection result code: {rc}'
-                       f' ({_get_mqtt_result(rc)})')
-    
+            _log.error(f'MQTT broker connection result code: {reason_code}'
+                       f' ({_get_mqtt_result(reason_code)})')
+
     def _mqtt_subscribe(self, topic: str, qos: int = 0):
         """Internal subscription handler assigns id indicating *subscribed*."""
         (result, mid) = self._mqtt.subscribe(topic=topic, qos=qos)
-        _log.debug(f'{self._client_id} subscribing to {topic}'
-                   f' (qos={qos}, mid={mid})')
+        if _vlog():
+            _log.debug(f'{self.client_id} subscribing to {topic}'
+                       f' (qos={qos}, mid={mid})')
         if result == MqttResultCode.SUCCESS:
             if mid == 0:
                 _log.warning(f'Received mid={mid} expected > 0')
@@ -295,29 +345,31 @@ class MqttClient:
             qos (int): The MQTT qos 0..2
 
         """
-        _log.debug(f'Adding subscription {topic} (qos={qos})')
+        if _vlog():
+            _log.debug(f'Adding subscription {topic} (qos={qos})')
         self._subscriptions[topic] = {'qos': qos, 'mid': 0}
         if self.is_connected:
             self._mqtt_subscribe(topic, qos)
         else:
-            _log.warning('MQTT not connected will subscribe later')
+            _log.debug(f'MQTT not connected...subscribing to {topic} later')
 
     def _mqtt_on_subscribe(self,
                            client: PahoClient,
-                           userdata: any,
+                           userdata: Any,
                            mid: int,
                            granted_qos: 'list[int]'):
         match = ''
-        for sub in self.subscriptions:
-            if mid == self.subscriptions[sub]['mid']:
-                _log.info(f'Subscription to {sub} successful'
+        for topic, detail in self.subscriptions.items():
+            if mid == detail.get('mid', None):
+                _log.info(f'Subscription to {topic} successful'
                           f' (mid={mid}, granted_qos={granted_qos})')
-                match = sub
+                match = topic
                 break
         if not match:
             _log.error(f'Unable to match mid={mid} to pending subscription')
 
-    def is_subscribed(self, topic: str):
+    def is_subscribed(self, topic: str) -> bool:
+        """Returns True if the specified topic is an active subscription."""
         if (topic in self.subscriptions and
             self.subscriptions[topic]['mid'] > 0):
             return True
@@ -325,22 +377,25 @@ class MqttClient:
 
     def _mqtt_on_message(self,
                          client: PahoClient,
-                         userdata: any,
+                         userdata: Any,
                          message: PahoMessage):
         """Internal callback on message simplifies passback to topic/payload."""
         payload = message.payload.decode()
         try:
             payload = json.loads(payload)
-        except json.JSONDecodeError as e:
-            _log.debug(f'MQTT message payload non-JSON ({e})')
-        _log.debug(f'MQTT received message "{payload}"'
-            f'on topic "{message.topic}" with QoS {message.qos}')
-        if userdata:
-            _log.debug(f'MQTT client userdata: {userdata}')
+        except json.JSONDecodeError as exc:
+            if _vlog():
+                _log.debug(f'MQTT message payload non-JSON ({exc})')
+        if _vlog():
+            _log.debug(f'MQTT received message "{payload}"'
+                       f' on topic "{message.topic}" with QoS {message.qos}')
+            if userdata:
+                _log.debug(f'MQTT client userdata: {userdata}')
         self.on_message(message.topic, payload)
 
     def _mqtt_unsubscribe(self, topic: str):
-        _log.debug(f'{self._client_id} unsubscribing to {topic}')
+        if _vlog():
+            _log.debug(f'{self.client_id} unsubscribing to {topic}')
         (result, mid) = self._mqtt.unsubscribe(topic)
         if result != MqttResultCode.SUCCESS:
             _log.error(f'MQTT Error {result} unsubscribing to {topic}'
@@ -353,38 +408,44 @@ class MqttClient:
             topic (str): The MQTT topic to unsubscribe
 
         """
-        _log.debug(f'Removing subscription {topic}')
+        if _vlog():
+            _log.debug(f'Removing subscription {topic}')
         if topic in self._subscriptions:
             del self._subscriptions[topic]
         if self.is_connected:
             self._mqtt_unsubscribe(topic)
 
-    def _mqtt_on_disconnect(self, client: PahoClient, userdata: any, rc: int):
+    def _mqtt_on_disconnect(self,
+                            client: PahoClient,
+                            userdata: Any,
+                            reason_code: int):
         """Internal callback when disconnected, clears subscription status."""
-        if self.on_disconnect:
-            self.on_disconnect(client, userdata, rc)
+        if callable(self.on_disconnect):
+            self.on_disconnect(client, userdata, reason_code)
         if userdata != 'terminate':
             _log.warning('MQTT broker disconnected'
-                         f' - result code {rc} ({_get_mqtt_result(rc)})')
-            self._mqtt.loop_stop()
-            for sub in self.subscriptions:
-                self._subscriptions[sub]['mid'] = 0
-            # get new unique ID to avoid bouncing connection
-            self.is_connected = False
-            if self.auto_connect:
-                if self._client_uid:
-                    self.client_id = self.client_id
-                self.connect()
+                         f' - result code {reason_code} ({_get_mqtt_result(reason_code)})')
+            # reconnect handling is managed automatically by Paho library
 
-    def publish(self, topic: str, message: 'str|dict', qos: int = 1) -> bool:
+    def publish(self,
+                topic: str,
+                message: 'str|dict|None',
+                qos: int = 1,
+                camel_keys: bool = False,
+                ) -> bool:
         """Publishes a message to a MQTT topic.
 
         If the message is a dictionary, 
         
         Args:
-            topic (str): The MQTT topic
-            message (str|dict): The message to publish
-            qos (int): The MQTT Quality of Service (0, 1 or 2)
+            topic: The MQTT topic
+            message: The message to publish
+            qos: The MQTT Quality of Service (0, 1 or 2)
+            camel_keys: Ensures all embedded dictionary keys are JSON style
+                (camelCase)
+
+        Returns:
+            True if successful, else False.
 
         """
         if message and not isinstance(message, (str, dict)):
@@ -393,28 +454,24 @@ class MqttClient:
             device_to_cloud = f'devices/{self.client_id}/messages/events/'
             if device_to_cloud not in topic:
                 _log.warning('Applying Azure device-to-cloud topic prefix')
-                topic = f'{device_to_cloud}/{topic}'
+                topic = f'{device_to_cloud}{topic}'
         if isinstance(message, dict):
-            for k, v in message.items():
-                if not isinstance(k, str):
-                    _log.warning(f'{k} ({type(k)}) will be converted to str(k)')
-                if isinstance(v, JSON_DUMPS_COMPATIBLE) or v is None:
-                    continue
-                if hasattr(v, '__dict__'):
-                    try:
-                        message[k] = vars(v)
-                    except TypeError:
-                        errstr = f'Could not convert [{k}] = {type(v)}'
-                        _log.warning(errstr)
-                        raise ValueError(f'Unhandled message...{errstr}')
-            message = json.dumps(message, skipkeys=True)
+            message = json.dumps(json_compatible(message, camel_keys),
+                                 skipkeys=True)
         if not isinstance(qos, int) or qos not in range(0, 3):
             _log.warning(f'Invalid MQTT QoS {qos} - using QoS 1')
             qos = 1
-        (rc, mid) = self._mqtt.publish(topic=topic, payload=message, qos=qos)
-        _log.debug(f'MQTT published (mid={mid}) {topic} {message}')
-        if rc != MqttResultCode.SUCCESS:
-            errmsg = f'Publishing error {rc} ({_get_mqtt_result(rc)})'
+        (reason_code, mid) = self._mqtt.publish(topic=topic,
+                                                payload=message,
+                                                qos=qos)
+        if _vlog():
+            _log.debug(f'MQTT published (mid={mid}) {topic} {message}')
+        if reason_code != MqttResultCode.SUCCESS:
+            errmsg = f'Publishing error {reason_code} ({_get_mqtt_result(reason_code)})'
             _log.error(errmsg)
             return False
         return True
+
+
+def _vlog() -> bool:
+    return verbose_logging('mqtt')
